// SDL WIKI - https://wiki.libsdl.org/
// Switch plasma demo by mrdude
#include <unistd.h>
#include <SDL.h>
#include <SDL_mixer.h>
//#include <SDL_image.h>
//#include <SDL_ttf.h>
#include <switch.h>
#include <string>
using namespace std;

#define RGB(r, g, b) (r << 16 | g << 8 | b)

const int kWidth = 320;
const int kHeight = 240;
int plasma = 1; // default plasma set to use routine 1 (black and white)
int plasma_num = 4; //set for max number of plasma effects
int maxsongs = 6;

string mysong = "data/lolas_space_fighter.mod"; //default music when app starts

Mix_Music *music = NULL;

float plasma1(float u, float v)
{
    Uint32 ticks = SDL_GetTicks();
    float time = ticks * 0.0025;
    float v1 = sinf(u * 10 + time);
    float v2 = sinf(25 * (u * sinf(time / 2) + v * cosf(time / 3)) + time);
    float cu = u + 0.5f * sinf(time / 5);
    float cv = v + 0.5f * cosf(time / 3);
    float v3 = sinf(sqrtf(200*(cu * cu + cv * cv) + 1) + time);
    return v1 + v2 + v3;
}

float plasma2(float u, float v)
{
    Uint32 ticks = SDL_GetTicks();
    float time = ticks * 0.0025;
    float v1 = sinf(16 * u + time);
    float v2 = sinf(32 * v + time / 3);
    float v3 = sinf((u + v) * 16 + time / 2);
    float v4 = sinf(32 * hypot(u, v) + time / 5);
    return (v1 + v2 + v3 + v4) / 4.0;
}

float plasma3(float u, float v)
{
    Uint32 ticks = SDL_GetTicks();
    float time = ticks * 0.0025;
    float v1 = sinf(u * 10 + time);
    float v2 = sinf(25 * (u * sinf(time / 2) + v * cosf(time / 3)) + time);
    float cu = u + 0.5f * sinf(time / M_PI);
    float cv = v + 0.5f * cosf(time / 3);
    float v3 = sinf(sqrtf(200*(cu * cu + cv * cv) + 1) + time);
    return v1 + v2 + v3;
}

float plasma4(float u, float v)
{
    Uint32 ticks = SDL_GetTicks();
    float time = ticks * 0.0025;
    float v1 = sinf(u * 10 + time);
    float v2 = sinf(25 * (u * sinf(time / 2) + v * cosf(time / 3)) + time);
    float cu = u + 0.5f * sinf(time / 5);
    float cv = v + 0.5f * cosf(time / 3);
    float v3 = sinf(sqrtf(200*(cu * cu + cv * cv) + 1) + time);
    return v1 + v2 + v3;
}

void draw(Uint32* pixels, int plasma)
{
    for (int x = 0; x < kWidth; ++x) {
        for (int y = 0; y < kHeight; ++y) {
            float u = -0.5f + float(x) / kWidth;
            float v = -0.5f + float(y) / kHeight;
            float gv = 0.0;
            
            //switch to change plasma routines
            switch (plasma) {
                case 1:
                    gv = plasma1(u, v); // / 3.0;
                    break;
                case 2:
                    gv = plasma2(u, v);
                    break;
                case 3:
                    gv = plasma3(u, v);
                    break;
                case 4:
                    gv = plasma4(u, v);
                    break;
            }
            
            if (plasma == 1){
            	//https://www.bidouille.org/prog/plasma 
            	//black and white
            	float random = (0.5 + static_cast <float> (rand()) / ( static_cast <float> (RAND_MAX/(5.0-0.5))));           	
            	Uint8 b = 255 * (sinf(gv * random * M_PI));
            	Uint8 r = b;
            	Uint8 g = r;
            	pixels[y * kWidth + x] = RGB(r, g, b);
            }
            
            if (plasma == 2){
            	Uint8 r = 255;
            	Uint8 g = 255 * (0.5f + 0.5f * cosf(gv * M_PI));
            	Uint8 b = 255 * (0.5f + 0.5f * sinf(gv * M_PI));
            	pixels[y * kWidth + x] = RGB(r, g, b);
            }
            
            if (plasma == 3){
            	Uint8 r = 255 * (0.5f + 1.2f * sinf(gv * M_PI));
            	Uint8 g = 255 * (0.8f + 0.5f * sinf(gv * M_PI));
            	Uint8 b = 255;
            	pixels[y * kWidth + x] = RGB(r, g, b);
            }
            
            if (plasma == 4){
            	Uint8 r = 255 * (0.8f + 1.2f * sinf(gv * M_PI));
            	Uint8 g = 255;
            	Uint8 b = 255 * (0.6f + 0.5f * sinf(gv * 2.5));
            	pixels[y * kWidth + x] = RGB(r, g, b);
            }
        }
    }
}

void musicselect(int songselection){
	switch (songselection) {
		case 1:
			mysong = "data/lolas_space_fighter.mod";
			break;
		case 2:
			mysong = "data/2ND_PM.S3M";
			break;
		case 3:
			mysong = "data/DEADLOCK.XM";
			break;
		case 4:
			mysong = "data/y_dptr.it";
			break;
		case 5:
			mysong = "data/invasion.xm";
			break;
		case 6:
			mysong = "data/delusion_16.xm";
			break;
  }
}


void play_music() {
	//https://lazyfoo.net/SDL_tutorials/lesson11/index.php
	Mix_Init(MIX_INIT_OGG);
	SDL_InitSubSystem(SDL_INIT_AUDIO);
	Mix_AllocateChannels(5);
	Mix_OpenAudio(48000, AUDIO_S16, 2, 4096);
	const char * c = mysong.c_str();
  music = Mix_LoadMUS(c);
	if (music)
		Mix_PlayMusic(music, -1);
}

int main(int argc, char* argv[])
{
	// Configure our supported input layout: a single player with standard controller styles
    padConfigureInput(1, HidNpadStyleSet_NpadStandard);

    // Initialize the default gamepad (which reads handheld mode inputs as well as the first connected controller)
    PadState pad;
    padInitializeDefault(&pad);
		
		romfsInit();
    chdir("romfs:/");
    
    bool quit = false;  
    //set screen up
    SDL_Init(SDL_INIT_VIDEO);
    SDL_Window* window = SDL_CreateWindow("PLASMA", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, 3 * kWidth, 3 * kHeight, 0);
    SDL_Renderer* renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_PRESENTVSYNC);
    SDL_Texture* texture = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_ARGB8888, SDL_TEXTUREACCESS_STATIC, kWidth, kHeight);
    Uint32* pixels = new Uint32[kWidth * kHeight];
    memset(pixels, 0, kWidth * kHeight * sizeof(Uint32));
    
    // load music and sounds from files
    play_music();
    bool music_toggle = false;
    
    int plasma = 1; // default plasma set to use routine 1 (black and white)
    int songselection = 1;
    
    //start drawing stuff on the screen
    while (!quit) {
    	draw(pixels, plasma);
    	SDL_UpdateTexture(texture, NULL, pixels, kWidth * sizeof(Uint32));
    	
    	// Scan the gamepad. This should be done once for each frame
    	padUpdate(&pad);
    	// padGetButtonsDown returns the set of buttons that have been newly pressed in this frame compared to the previous one
    	u64 kDown = padGetButtonsDown(&pad);
    	
    	if (kDown & HidNpadButton_Plus){
    		quit = true;
    	}
    	
    	if (kDown & HidNpadButton_A){
    		plasma = (plasma + 1);
    		if (plasma > plasma_num){
    			plasma = 1; //just go back to start...
    		}
    	}
    	
    	if (kDown & HidNpadButton_B){
    		plasma = (plasma - 1);
    		if (plasma < 1){
    			plasma = plasma_num; //we are just basically looping the effect
    		}
    	}
    	
    	if (kDown & HidNpadButton_X){ // change music
    		songselection = (songselection + 1);
    		if (songselection > maxsongs){
    			songselection = 1;
    		}
    		musicselect(songselection);
    		Mix_FreeMusic(music);
    		play_music();   		
    	}
    	
    	if (kDown & HidNpadButton_Y){ // pause music
    		if (!music_toggle){
    			Mix_PauseMusic();
    			music_toggle = true;
    		}
    		
    		else {
    			music_toggle = false; // resume music
    			Mix_ResumeMusic();
    		}
    	}
    	
    	SDL_RenderClear(renderer);
    	SDL_RenderCopy(renderer, texture, NULL, NULL);
    	SDL_RenderPresent(renderer);
    }

    //when we exit out of the loop - clean up and quit the program
    delete[] pixels;
    
    //Free the music
    Mix_FreeMusic(music);
    //Quit SDL_mixer
    Mix_CloseAudio();
    
    SDL_DestroyTexture(texture);
    SDL_DestroyRenderer(renderer);
    SDL_DestroyWindow(window);
    SDL_Quit();
    romfsExit();
    return 0;
}